require("util")

local function toChunk(position)
    return {x = math.floor(position.x) / 32, y = math.floor(position.y) / 32} 
end

local function isPositionInList(position, listOfPositions)
    for _, pos in pairs(listOfPositions) do 
        if pos.x == position.x and pos.y == position.y then 
            return true 
        end 
    end
    return false
end

local function expandOutwardsByOne(positions)
    local seenTiles = {}
    local res = {}
    for _,tile in pairs(positions) do
        if seenTiles[tostring(tile.x+32)..":"..tostring(tile.y)] == nil then table.insert(res, {x=tile.x+32,y=tile.y}) seenTiles[tostring(tile.x+32)..":"..tostring(tile.y)] = true end
        if seenTiles[tostring(tile.x-32)..":"..tostring(tile.y)] == nil then table.insert(res, {x=tile.x-32,y=tile.y}) seenTiles[tostring(tile.x-32)..":"..tostring(tile.y)] = true end
        if seenTiles[tostring(tile.x)..":"..tostring(tile.y+32)] == nil then table.insert(res, {x=tile.x,y=tile.y+32}) seenTiles[tostring(tile.x)..":"..tostring(tile.y+32)] = true end
        if seenTiles[tostring(tile.x)..":"..tostring(tile.y-32)] == nil then table.insert(res, {x=tile.x,y=tile.y-32}) seenTiles[tostring(tile.x)..":"..tostring(tile.y-32)] = true end
    end
    return res
end

local function expandOutwards(positions, numTiles)
    local res = positions
    for i = 1, numTiles do
        res = expandOutwardsByOne(res)
    end
    return res
end

local function init_globals()
    
    --global.generator = game.create_random_generator()
    global.islandCandidates = {} --list of positions {x,y}
    global.tilesOfIsland = {} --list of positions {x,y}
    global.expandedTilesOfIsland = {} --list of positions {x,y}
    global.notTilesOfIsland = {} --list of positions {x,y}
    global.remainingTilesOfIslandToCheck = {} --list of positions {x,y}
    --global.deleteSurfaceNextTick = false
    --global.islandNumer = 0
    
    global.version = "0.1.1"
                
end

script.on_init(function()
    init_globals()
end)

local function updateModVersion()
    --nothing to do yet
end

script.on_configuration_changed(function()
    updateModVersion()
end)

script.on_event(defines.events.on_tick, function(event)

    --local islandSurfaceName = "island-back-buffer"..tostring(global.islandNumer)
    local islandSurfaceName = "island-back-buffer"
    local islandSurface = game.surfaces[islandSurfaceName]
    local normalSurfaceName = "nauvis"
    local normalSurface = game.surfaces[normalSurfaceName]
        
    if islandSurface == nil then
        
        game.create_surface(islandSurfaceName, {
            terrain_segmentation=normalSurface.map_gen_settings.terrain_segmentation,
            water=normalSurface.map_gen_settings.water,
            autoplace_controls = normalSurface.map_gen_settings.autoplace_controls,
            --seed = (normalSurface.map_gen_settings.seed + game.tick) % 4294967296,
            seed = (normalSurface.map_gen_settings.seed + math.random(0,4294967295)) % 4294967296,
            --width = 10000, --?
            --height = 10000, --?
            starting_area = normalSurface.map_gen_settings.starting_area,
            starting_points = normalSurface.map_gen_settings.starting_points, --??? maybe this one can be used to trigger multiple islands from the start!!
            peaceful_mode = normalSurface.map_gen_settings.peaceful_mode,
            research_queue_from_the_start = normalSurface.map_gen_settings.research_queue_from_the_start,
            property_expression_names = normalSurface.map_gen_settings.property_expression_names
        })
        
        game.surfaces[islandSurfaceName].request_to_generate_chunks({x=0,y=0},10)
        
        table.insert(global.remainingTilesOfIslandToCheck, {x=0,y=0})
        return
    end
    
    if #global.remainingTilesOfIslandToCheck > 0 then
        local tileToCheck = table.remove(global.remainingTilesOfIslandToCheck)
        game.players[1].print("#tilesOfIsland: " .. tostring(#global.tilesOfIsland) .. " #notTilesOfIsland " .. tostring(#global.notTilesOfIsland))
        if not islandSurface.is_chunk_generated(toChunk(tileToCheck)) then
            islandSurface.request_to_generate_chunks(tileToCheck)
            islandSurface.force_generate_chunk_requests()
            table.insert(global.remainingTilesOfIslandToCheck, tileToCheck) --put it back, try later
            return
        end
        if islandSurface.count_tiles_filtered{area={{tileToCheck.x, tileToCheck.y}, {tileToCheck.x+32, tileToCheck.y+32}}, name="deepwater"} == 32*32 then
            --pure water
            table.insert(global.notTilesOfIsland, tileToCheck)
        else
            table.insert(global.tilesOfIsland, tileToCheck)
            if not isPositionInList({x=tileToCheck.x + 32,y=tileToCheck.y}, global.tilesOfIsland) and not isPositionInList({x=tileToCheck.x + 32,y=tileToCheck.y}, global.notTilesOfIsland) 
                then table.insert(global.remainingTilesOfIslandToCheck, {x=tileToCheck.x + 32,y=tileToCheck.y}) end
            if not isPositionInList({x=tileToCheck.x - 32,y=tileToCheck.y}, global.tilesOfIsland) and not isPositionInList({x=tileToCheck.x - 32,y=tileToCheck.y}, global.notTilesOfIsland) 
                then table.insert(global.remainingTilesOfIslandToCheck, {x=tileToCheck.x - 32,y=tileToCheck.y}) end
            if not isPositionInList({x=tileToCheck.x,y=tileToCheck.y + 32}, global.tilesOfIsland) and not isPositionInList({x=tileToCheck.x,y=tileToCheck.y + 32}, global.notTilesOfIsland) 
                then table.insert(global.remainingTilesOfIslandToCheck, {x=tileToCheck.x,y=tileToCheck.y + 32}) end
            if not isPositionInList({x=tileToCheck.x,y=tileToCheck.y - 32}, global.tilesOfIsland) and not isPositionInList({x=tileToCheck.x,y=tileToCheck.y - 32}, global.notTilesOfIsland) 
                then table.insert(global.remainingTilesOfIslandToCheck, {x=tileToCheck.x,y=tileToCheck.y - 32}) end
        end
        return
    end
    
    if #global.expandedTilesOfIsland == 0 then
        global.expandedTilesOfIsland = expandOutwards(global.tilesOfIsland, math.random(2,4)) --todo settings with min max?
        return
    end
    
    if #global.islandCandidates == 0 then return end
    
    local islandCandidate = table.remove(global.islandCandidates)
    
    for _, tile in pairs(global.expandedTilesOfIsland) do
        if normalSurface.is_chunk_generated(toChunk({x=tile.x+islandCandidate.x,y=tile.y+islandCandidate.y})) then
            if normalSurface.count_tiles_filtered{area={{tile.x+islandCandidate.x, tile.y+islandCandidate.y}, {tile.x+32+islandCandidate.x, tile.y+32+islandCandidate.y}}, name="deepwater"} ~= 32*32 then
                game.players[1].print(tostring(game.tick).. "discarded candidate: " .. tostring(islandCandidate.x) .. " : " .. tostring(islandCandidate.y) .. " remaining: " .. tostring(#global.islandCandidates))
                return --discard candidate as there is no space on normal surface to place this island here
            end
        end
    end

    game.players[1].print(tostring(game.tick).. "final #tilesOfIsland: " .. tostring(#global.tilesOfIsland) .. " #notTilesOfIsland " .. tostring(#global.notTilesOfIsland))
    game.players[1].print(tostring(game.tick).. "accepted candidate: " .. tostring(islandCandidate.x) .. " : " .. tostring(islandCandidate.y) .. " remaining: " .. tostring(#global.islandCandidates))

    for _,tile in pairs(global.tilesOfIsland) do
        islandSurface.clone_area{
            source_area={{tile.x,tile.y},{tile.x+32, tile.y+32}}, 
            destination_area={{islandCandidate.x+tile.x,islandCandidate.y+tile.y},{islandCandidate.x+tile.x+32,islandCandidate.y+tile.y+32}},
            destination_surface=normalSurfaceName,
            clone_tiles=true,
            clone_entities=true,
            clone_decoratives=true,
            clear_destination=true,
            expand_map=true
        }
        normalSurface.regenerate_entity(nil, {toChunk(tile)})
        for _,force in pairs(game.forces) do
            if force.is_chunk_charted(normalSurface, toChunk(tile)) then force.chart(normalSurface, {{islandCandidate.x+tile.x,islandCandidate.y+tile.y},{islandCandidate.x+tile.x+32,islandCandidate.y+tile.y+32}}) end
        end
        
    end
    global.tilesOfIsland = {}
    global.notTilesOfIsland = {}
    global.expandedTilesOfIsland = {}
    game.delete_surface(islandSurface)
    --global.islandNumer = global.islandNumer + 1
    normalSurface.force_generate_chunk_requests()
end)

script.on_event(defines.events.on_chunk_generated, function(e)

    if e.surface == game.surfaces["nauvis"] then
        
        local chunk = e.area
        local chunkIdString = tostring(chunk.left_top.x) .. ":" .. tostring(chunk.left_top.y)
        if math.random(0,9) == 0 then
            table.insert(global.islandCandidates, chunk.left_top)
            --game.players[1].print("added candidate " .. tostring(chunk.left_top.x) .. " : " .. tostring(chunk.left_top.y))
        end
    end    
end)